/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef PYRED_PARSER_H
#define PYRED_PARSER_H

#include <string>
#include <vector>
#include <stack>
#include <unordered_map>

#include "coeff_helper.h" // icpair

namespace pyred
{

template<typename Coeff>
using arg_stack_t = std::stack<Coeff,std::vector<Coeff>>;

bool is_left_assoc(const char);
int op_priority_diff(const char, const  char);
bool is_operator(const char);
std::vector<std::string> split(const std::string& s, char delim, int n = 0);
std::pair<intid,std::string> split_intcoeff(const std::string&);
std::vector<std::string> tokenize(const std::string&);

template<typename Coeff>
class ParserCache {
  using cache_type = std::unordered_map<std::string,Coeff>;
public:
  static cache_type cache;
  static typename cache_type::const_iterator find(const std::string& s)
  {
    return cache.find(s);
  }
  static bool insert(const std::string& key, const Coeff& val) {
    return cache.insert({key, val}).second;
  }
  static typename cache_type::const_iterator cend() {
    return cache.cend();
  }
  static std::size_t size() {return cache.size();}
  static void clear() {cache.clear();}
};
template<typename Coeff>
typename ParserCache<Coeff>::cache_type ParserCache<Coeff>::cache{};


template<typename Coeff>
void stack_eval(arg_stack_t<Coeff>& arg_stack, const char op) {
  if (arg_stack.empty()) {
    throw parser_error("operator is lacking arguments");
  }
  Coeff rarg = arg_stack.top();
  arg_stack.pop();
  switch (op) {
    case '+':
      arg_stack.top() += rarg;
      break;
    case '-':
      arg_stack.top() -= rarg;
      break;
    case '!': // unary '-'
      arg_stack.emplace(-rarg);
      break;
    case '*':
      arg_stack.top() *= rarg;
      break;
    case '/':
      arg_stack.top() /= rarg;
      break;
    case '^':
      arg_stack.top() = std::move(pow(arg_stack.top(), rarg));
      break;
    default:
      throw parser_error(std::string("invalid operator ") + std::string(1,op));
  }
  return;
}

template<typename Coeff>
Coeff parse_coeff(const std::string& expr) {
  /* Shunting-yard algorithm with on-the-fly evaluation.
     Parenthesis are not supported.
  */
  // cache lookup
  const auto cit = ParserCache<Coeff>::find(expr);
  if (cit != ParserCache<Coeff>::cend()) {
    return cit->second;
  }
  arg_stack_t<Coeff> arg_stack;
  std::stack<char,std::vector<char>> op_stack;
  std::vector<std::string> tokens;
  try {
    tokens = tokenize(expr);
  }
  catch(const parser_error& e) {
    throw parser_error(e.msg + std::string(" in ") + expr);
  }
  for (const auto& token: tokens) {
    auto curop = token.front();
    if (is_operator(curop)) {
      if (!op_stack.empty()) {
        auto prevop = op_stack.top();
        // While the token on top of op_stack is an operator prevop and
        // * curop is left-associative
        //   and its precedence is the same as that of prevop
        // * or curop has lower precedence than prevop:
        while (is_operator(prevop) && (
               (is_left_assoc(curop) && op_priority_diff(curop, prevop) == 0)
               || (op_priority_diff(curop, prevop) < 0))) {
          stack_eval<Coeff>(arg_stack, prevop);
          op_stack.pop();
          if (!op_stack.empty()) {
            prevop = op_stack.top();
          }
          else {
            break;
          }
        }
      }
      // push token onto the operator op_stack.
      op_stack.push(curop);
    }
    else if (curop == '(') {
      op_stack.push(curop);
    }
    else if (curop == ')') {
      // push operators from op_stack onto arg_stack (i.e. evaluate)
      // until a left parenthesis is encountered.
      while (op_stack.top() != '(') {
        stack_eval<Coeff>(arg_stack, op_stack.top());
        op_stack.pop();
        if (op_stack.empty()) {
          // op_stack ran empty without encountering
          // a matching left parenthesis
          throw parser_error(
            std::string("mismatched parenthesis in ") + expr);
        }
      }
      // pop the left parenthesis from op_stack
      if (!op_stack.empty()) op_stack.pop();
    }
    else {
      // token is an argument, i.e. a number or a symbol
      arg_stack.push(token);
    }
  }
  while (!op_stack.empty()) {
    // apply operators from op_stack to arguments
    if (op_stack.top() == '(' || op_stack.top() == ')') {
      throw parser_error(std::string("mismatched parenthesis in ") + expr);
    }
    stack_eval<Coeff>(arg_stack, op_stack.top());
    op_stack.pop();
  }
  // fill the cache
  ParserCache<Coeff>::insert(expr, arg_stack.top());
  return arg_stack.top();
}

}

#endif
